// -*- tab-width:4 -*-
#pragma once
#if !defined(ALUTGUI_HPP_)
#define ALUTGUI_HPP_

/**@file
   GUI関連の補助的なユーティリティ。
 **/

#include <string>
#include <sstream>
#include <map>
#include <set>
#include <vector>

#include <iostream>

#include <WSCbase.h>
#include <WSDkeyboard.h>
#include <WSCfontSet.h>
#include <WSKeysym.h>

//#define ALUTGUI_RECTANGLE_DEBUG
#if defined(ALUTGUI_RECTANGLE_DEBUG)
#include "ALAppLog.hpp"
#endif

#include "ALUtString.hpp"

class WSCvifield;

namespace al {
namespace gui {

/** アプリケーションのメインウィンドウを登録する。 **/
void SetAppMainWindow(WSCbase* window);

/** アプリケーションのメインウィンドウを取得する。 **/
WSCbase* GetAppMainWindow();

/** メッセージボックスのタイトル文字列のデフォルトを登録する。 */
void SetAppTitleString(const std::string& app_title_string);

/** メッセージボックスのタイトル文字列のデフォルトを取得する。 */
const std::string GetAppTitleString();


/** ユーザー操作モード */
enum ActionCode { 
	AC_OK=1,					/**< 一般的な肯定応答 */
	AC_CANCEL=2,				/**< 一般的な取り消し応答 */
	AC_YES=4,					/**< 質問に対する肯定応答 */
	AC_NO=8,					/**< 質問に対する否定応答 */
    AC_ERROR=0                  /**< 処理でエラーが発生したことを示す。（戻り値としてのみ使われる。）*/
};

const char *ActionCodeStr(ActionCode code);

/** 画面にメッセージパネルを表示して、ユーザーの反応を返す。
	（ログ出力無し）*/
ActionCode PopupMsgBox(
	const std::string&	msg,
	int					action_code_set,
    const std::string&	title="",
    WSCbase*            parent=NULL
	);

/** 画面にメッセージパネルを表示して、ユーザーの反応を返す。
	（ログ出力有り）*/
ActionCode MsgBox(
	const std::string&	msg,
								/**< 表示するメッセージ（UTF-8）*/
	int					action_code_set	= (AC_OK|AC_CANCEL),
								/**< 表示するボタン（ActionCodeの論理和で示す）*/
    const std::string&	title="",
    WSCbase*            parent=NULL
	);

/** 画面にメッセージパネルを表示して、ユーザーからの入力文字列を受け取る。
	（ログ出力無し）*/
ActionCode PopupInputBox(
	const std::string&	msg,
	std::string&		input_text,
    WSCbase*            parent=NULL
	);

/** 画面にメッセージパネルを表示して、ユーザーからの入力文字列を受け取る。
	（ログ出力有り）*/
ActionCode InputBox(
	const std::string&	msg,
	std::string&		input_text,
    WSCbase*            parent=NULL
	);

/** 親ウィンドウの上へウィンドウを配置する。
    （WSNx/WSNyが未設定の時のMWTの動きが腑に落ちないので、確定するようにする）
 */
void LocateWindowOnParent( WSCbase* parent, WSCbase* win);

/** 画面へログ由来のエラーメッセージがpopup表示された場合に、追加表示するメッセー
	ジテキストを管理するクラス兼唯一のインスタンス。 */
class ErrorPopup {
	std::vector<std::string>	m_msg;

	ErrorPopup() {}
	ErrorPopup(const ErrorPopup&);
	void operator=(const ErrorPopup&);
public:
	void push(const std::string& msg) { m_msg.push_back(msg); }
	void pop() { m_msg.pop_back(); }

	static ErrorPopup *GetInst() {
		static ErrorPopup inst;
		return &inst;
	}

	static bool PopupRequested() {
		if( GetInst()->m_msg.empty() )
			return true;
		return (GetInst()->m_msg.back() != "NoPopup");
	}
	
	static const std::string GetLastErrorPopup() {
		if( GetInst()->m_msg.empty() )
			return "";
		return GetInst()->m_msg.back();
	};
};


/** エラー時に表示されるポップアップパネルに追加表示するメッセージを管理する。 */
class RegisterErrorPopup {
	RegisterErrorPopup();
public:
	RegisterErrorPopup(const std::string& msg) { 
		ErrorPopup::GetInst()->push(msg);
	}
	~RegisterErrorPopup() {
		ErrorPopup::GetInst()->pop();
	}
};

/** 指定のフォントで文字列を表示した際の幅と高さを求める */
void GetTextExtent(
	const std::string& str,
	long& width,
	long& height,
	WSDfont* font = NULL
    );

/** 指定のフォントで文字列を表示した際の幅を求める。<br>
    フォント指定が無い時はデフォルトフォントを使う。 */
unsigned short GetTextWidth(
	const std::string& str,
	WSDfont* font = NULL
    );

/** 指定のフォントで文字列を表示した際の高さを求める。<br>
    フォント指定が無い時はデフォルトフォントを使う。 */
unsigned short GetTextHeight(
	const std::string& str,
	WSDfont* font = NULL
    );

/** 指定のオブジェクトで使われる WSDfontへのポインタを返す。 */
WSDfont* GetObjectFont(WSCbase* object);

/** 文字列A（UTF-8）の各文字間に、文字列Bを挟み込んだものを返す */
std::string InsertStubs(const std::string& str_a, const std::string& str_b );

/** 入力フィールドのフォーカス状態に変化があったときに呼び出す。 */
void OnInputFieldFocusChanged( WSCbase* object );

/** フォーカスがある入力フィールドを設定する。 */    
void SetFocusedInputField(WSCbase* object);

/** フォーカスがある（厳密には直前までフォーカスがあった）入力フィールドを返す */    
WSCbase* GetFocusedInputField();

/** パーツとして使われている vifield へのポインタを返す。  */
WSCvifield* Get_vifield(WSCbase* field);

/** Undoする為の情報を確保する。 */
void KeepUndoStatusForInputField(WSCbase* field);

/** Undoする為の情報を確定する。 */
void CheckUndoStatusForInputField(WSCbase* field);

/** クリップボードとT型の入力フィールド間でコピー／カット処理を行なう */
template<typename T>
bool CopyToClipboard(T* ifield, bool cut=false)
{
    if( ifield == NULL || ifield->getVisible() != True )
        return false;

    WSCstring	str = ifield->getSelectedString();
    WSDkeyboard*	keyboard=WSGIappKeyboard();
    keyboard->setSelectedString(str);
    if( cut ) {
        KeepUndoStatusForInputField(ifield);
        ifield->deleteSelectedString();
        CheckUndoStatusForInputField(ifield);
        ifield->setFocus();
    }
    return true;
}

/** クリップボードとT型の入力フィールド間でペースト処理を行なう */
template<typename T>
bool PasteFromClipboard(T* ifield)
{
    if( ifield == NULL
        || ifield->getVisible() != True
        || ifield->getSensitive() != True
        || static_cast<WSCbool>(ifield->getProperty(WSNdisplayOnly)) != False )
        return false;

    WSDkeyboard*	keyboard=WSGIappKeyboard();
    WSCstring	str = keyboard->getSelectedString();
    KeepUndoStatusForInputField(ifield);
    ifield->replaceSelectedString(str);
    CheckUndoStatusForInputField(ifield);
    ifield->setFocus();

    return true;
}

/** objectがサポートされている入力フィールドであれば、
    クリップボードとの間でコピー処理を行なう。*/
bool CopyFromInputFieldToClipboard(WSCbase* object, bool cut=false);

/** objectがサポートされている入力フィールドであれば、
    クリップボードとの間でペースト処理を行なう。*/
bool PasteFromClipboardToInputField(WSCbase* object);

/** 入力フィールドがコピー可能もしくは編集可能かの状態を調べてフラグをセットする。
    （型直接指定テンプレート版）*/
template<typename TInput>
void GetInputFieldConditionForEditing(
    TInput* field,
    bool& copy_available,
    bool& edit_available,
    bool& paste_available)
{
    copy_available = false;
    edit_available = false;
    paste_available = false;
    if( field != NULL ) {
        if( field->getSelectedString().getChars() != 0 )
            copy_available = true;
        WSCbool	display_only = field->getProperty(WSNdisplayOnly);
        if( !display_only && field->getSensitive() ) {
            edit_available = true;

            WSDkeyboard*	keyboard=WSGIappKeyboard();
            WSCstring	str = keyboard->getSelectedString();
            if( str.getChars() != 0 )
                paste_available = true;
        }
    }
}

/** 入力フィールドがコピー可能もしくは編集可能かの状態を調べてフラグをセットする。
    （型推定ポリモルフィズム版） */
void CheckInputFieldConditionForEditing(
    WSCbase* input_field,
    bool& copy_available,
    bool& edit_available,
    bool& paste_available);


/** オブジェクトとインスタンスを結びつけるための仕掛け。<br>
    イベントプロシージャ側からクラスインスタンスをアクセスする必要がある場合を想定。<br>
    addProcedure する場所で Memorizeしておき、イベントプロシージャ側で Recall する。<br>
    他にも使い道はあるだろう。

    @attention
    このテンプレートから実体化されるインスタンスは、イベントプロシージャから呼び
    出せるようになっている必要がある。したがって、クラス宣言の中でこのテンプレー
    トを実体化するときは、static メンバ変数にする。ファイルローカル変数として実体
    化する方法もある。
     */

template<typename Tobject, typename Tinstance>
class ObjectInstanceMap {

    std::map<Tobject*,Tinstance*>	m_object_instance_map;

public:
    void MemorizeInstanceForTheObject(Tobject *object,Tinstance* instance) {
        typename std::map<Tobject*,Tinstance*>::value_type	value(object,instance);
        m_object_instance_map.insert(value);
    }

    Tinstance* RecallInstanceForTheObject(Tobject *object) {
        typename std::map<Tobject*,Tinstance*>::iterator it;
        if( (it = m_object_instance_map.find(object)) == m_object_instance_map.end() )
            return NULL;
        return it->second;
    }

    void RemoveInstanceForTheObject(Tobject *object) {
        typename std::map<Tobject*,Tinstance*>::iterator it;
        if( (it = m_object_instance_map.find(object)) != m_object_instance_map.end() )
            m_object_instance_map.erase(it);
    }

    void RemoveMappingsForTheInstance(Tinstance *instance) {
        typename std::map<Tobject*,Tinstance*>::iterator it;
        typename std::map<Tobject*,Tinstance*>::iterator it_last;

        it = m_object_instance_map.begin();
        while( it != m_object_instance_map.end() ) {
            if( it->second == instance )
                m_object_instance_map.erase(it++);
            else
                ++it;
        }
    }

    void Clear() {
        m_object_instance_map.clear();
    }
};

/**
   オブジェクトを識別するための情報を一元管理するためのクラス。
 **/
template<typename Tobject>
class ObjectCollectionManager :
        public std::map<const std::string,Tobject> {
public:
    ObjectCollectionManager() {}
    ~ObjectCollectionManager() {}

};

/**
   ウィンドウパーツを識別するための情報を一元管理するためのコレクション型。
 **/
typedef ObjectCollectionManager<std::vector<WSCbase *> > WindowPartsCollection;

/**
   WindowPartsCollection 操作の為のペア
 **/
typedef std::pair<std::string,std::vector<WSCbase *> > WindowPartsCollPair;

/**
   ウィンドウを識別するための情報を一元管理するためのコレクション型。
 **/
typedef ObjectCollectionManager<WindowPartsCollection> WindowCollection;

/**
   WindowCollection 操作の為のペア
 **/
typedef std::pair<std::string,WindowPartsCollection> WindowCollPair;


bool GetWindowRectangle(WSCbase* window, int& x, int& y, int& w, int& h );

/**
   指定のウィンドウが指定のスクリーン座標にあるかを判定する。
 **/
bool WindowIsUnderThePoint(WSCbase* window, int screen_x, int screen_y); 

/**
   メッセージパネルを表示して、key_for_set/key_for_getいずれかのキー入力を待ち、
   キーが押された時点のマウス座標と入力されたキーのキーコードを返す。<br>
   尚、マウス左ボタンクリック時は key_for_set、
   右ボタンクリック時は key_for_get が返される。<br>
   ESC 入力は常に有効。
 **/
long GetMousePosition(
    WSCbase* parent,
    const std::string& msg,
    int& screen_x,
    int& screen_y,
    long key_for_set=WSK_Return,
    long key_for_get=WSK_Insert );

/**
   位置管理の為のクラス。
 **/
struct Position {
    Position() : _x(0), _y(0) {}
    Position(long x, long y) : _x(x), _y(y) {}

    inline void Set( long x, long y ) { _x = x; _y = y; }
    inline void SetX( long x ) { _x = x; }
    inline void SetY( long y ) { _y = y; }
    inline long X() const { return _x; }
    inline long Y() const { return _y; }

private:
    long	_x;
    long	_y;
};

/**
   矩形管理のためのクラス。
 **/
struct Rectangle {

    Rectangle() : _x(0), _y(0), _w(0), _h(0) {}
    Rectangle(long x, long y, long w, long h ) { Set(x,y,w,h); }

    inline void Set( long x, long y, long w, long h ) {
        _x = x; _y = y; _w = w; _h = h;
    }
    inline void SetPos( long x, long y ) { _x = x; _y = y; }
    inline void SetWidth( long w ) { _w = w; }
    inline void SetHeight( long h ) { _h = h; }

    inline void SetLeft(long left, bool hold_width=true ) {
        if( hold_width ) _x = left; else _w = Right() - left+1;
    }
    inline void SetTop( long top, bool hold_height=true ) {
        if( hold_height ) _y = top; else _h = Bottom()-top+1;
    }
    inline void SetRight( long right, bool hold_width=true ) {
        if( hold_width ) _x = right-Width()+1; else _w = right-Left()+1;
    } 
    inline void SetBottom( long bottom, bool hold_height=true ) {
        if( hold_height ) _y = bottom-Height()+1; else _h = bottom-Top()+1;
    }

    inline long Left() const { return _x; }
    inline long Top() const { return _y; }
    inline long Right() const { return _x+_w-1; }
    inline long Bottom() const { return _y+_h-1; }
    inline long Width() const { return _w; }
    inline long Height() const { return _h; }
    inline Position TopLeft() const { return Position(Left(),Top()); }
    inline Position BottomRight() const { return Position(Right(),Bottom()); }

    void GetWindowRectangle(WSCbase*);

    inline bool IsUnderThePoint( int x, int y ) const {
        bool tobe = (Left() <= x && x <= Right()
                     && Top() <= y && y <= Bottom());

#if defined(ALUTGUI_RECTANGLE_DEBUG)
        ::al::applog::OutputAppLog(
            ::al::applog::LOG_ONLY,
            "Rectangle::IsUnderThePoint: (%d,%d)-(%d,%d) is%s under (%d,%d)",
            Left(), Top(), Right(), Bottom(), (tobe ? "" : " not"), x, y );
#endif
        return tobe;
    }

    inline bool IsInside(const Rectangle& rect) const {
        bool tobe = (rect.Left() <= Left() && Right() <= rect.Right()
                     && rect.Top() <= Top() && Bottom() <= rect.Bottom());

#if defined(ALUTGUI_RECTANGLE_DEBUG)
        ::al::applog::OutputAppLog(
            ::al::applog::LOG_ONLY,
            "Rectangle::IsInside: (%d,%d)-(%d,%d) is%s inside (%d,%d)-(%d,%d)",
            Left(), Top(), Right(), Bottom(), (tobe ? "" : " not"),
            rect.Left(), rect.Top(), rect.Right(), rect.Bottom() );
#endif
        return tobe;
    }


private:
    long _x;
    long _y;
    long _w;
    long _h;
};


/**
   ウィンドウの強制的な再描画。
 **/
void RedrawWindow(WSCbase* window,bool compulsory=false);

/**
   ウィンドウ群の強制的な再描画。
 **/
void RedrawWindows(const std::vector<WSCbase*>& windows, bool compulsory=false);

/** RGB個別設定用の定数 */
enum RGBId {RGB_R,RGB_G,RGB_B,MAX_RGB};

/** RGBピクセル値表現型 */
typedef unsigned long RGBPixelValueT;

/** RGB個別輝度値表現型 */
typedef unsigned char RGBColorT;

/** 三原色配列表現型 */
struct RGBColorArray {
    RGBColorT	m_value[MAX_RGB];
    RGBColorArray() { for( size_t i=0; i < MAX_RGB; ++i ) m_value[i]=0; }
};

/** RGB値を合成する。 */
inline RGBPixelValueT ComposeRGBPixelValue(
    unsigned int r, unsigned int g, unsigned int b) {
    return static_cast<RGBPixelValueT>(((r&0xff)<<16) | ((g&0xff)<<8) | (b&0xff));
}

/** RGB値を分解する。 */
inline void DecomposeRGBPixelValue(
    RGBPixelValueT pixval, unsigned int& r, unsigned int& g, unsigned int& b) {
    r = (pixval>>16)&0xff;
    g = (pixval>>8)&0xff;
    b = pixval&0xff;
}

/**
   数値形式の RGB値を#RRGGBB 文字列形式に変換。
 **/
const std::string RGBString(RGBPixelValueT rgb);

/**
   数値配列形式の RGB値を#RRGGBB 文字列形式に変換。
 **/
const std::string RGBString(const RGBColorT *rgb);

/**
   RGBPixelColorArray構造体形式の RGB値を#RRGGBB 文字列形式に変換。
 **/
const std::string RGBString(const RGBColorArray& rgb_array);

/**
   文字列形式の RGB値を数値形式に変換。
 **/
RGBPixelValueT RGBStringToRGBColors(const std::string& rgb_str);

/**
   文字列形式の RGB値を配列形式に変換。
 **/
void RGBStringToRGBColors(const std::string& rgb_str, RGBColorT *rgb);

/**
   文字列形式の RGB値を RGBPixelColorArray構造体形式に変換。
 **/
void RGBStringToRGBColorArray(const std::string& rgb_str, RGBColorArray& rgb_array);

/**
   ウィンドウに色関連のプロパティをセットする。<br>
   @note 下記の付加処理を行なう。
   <ul><li>背景色をセットする時は影の色を自動生成する。</li></ul>
 **/
void SetWindowColorProperty(WSCbase* window,
                            const std::string& propname, const std::string& propvalue);

/**
   保存／復元の定型的なメニューの生成処理（単位処理）。

   @attention カスケードメニューではアクセラレータは効かないみたいである。
 **/
template<typename WSCMenuType>
void MakeAttributeMenuString(
    const std::vector<std::string>&	tag_enum,
    const std::string&              menu_str,
    const std::string&              ep_str,
    int                             new_id,
    WSCMenuType*					menu,
    std::vector<std::string>&       submenu_str,
    const std::string&				icon_name=""
    )
{
    unsigned char font_no = menu->getProperty(WSNfont);
    WSDfont* font = WSGIappFontSet()->getFont(font_no);
    const std::string	space = " ";	// きっちりだとメニューの右端が切れる
    size_t	space_width = GetTextWidth(space,font);

/** @attention メニューの一番上の項目の長さがメニューの幅になるらしく、
    文字列長にあまり長さの差が発生しないように空白文字を使って調整する。
**/
    size_t	max_width = 0;
    for( size_t i = 0; i < tag_enum.size(); ++i ) {
        size_t width = GetTextWidth(tag_enum[i],font);
        if( max_width < width )
            max_width = width;
    }
    max_width += space_width;

    if( max_width == 0 || space_width == 0 )
        return;                 /* bug! */

    submenu_str.clear();

    if( tag_enum.size() == 0 && new_id < 0) {
        menu->addItem(const_cast<char*>(menu_str.c_str()),"","",0,
                      const_cast<char*>(icon_name.c_str()));
    }
    else {
        menu->beginCascade(const_cast<char*>(menu_str.c_str()));
        for( size_t i = 0; i < tag_enum.size(); ++i ) {
            std::stringstream	ss;
            ss << " ";
            ss << tag_enum[i];
            for( size_t w = GetTextWidth(tag_enum[i],font); w < max_width; w += space_width )
                ss << space;
            std::string sstr = ss.str();
            menu->addItem(const_cast<char*>(sstr.c_str()),
                          const_cast<char*>(ep_str.c_str()),
                          "",i,"");
            submenu_str.push_back(sstr);
        }
        if( new_id >= 0 )
            menu->addItem("新しい名前で保存 ... ",
                          const_cast<char*>(ep_str.c_str()),"",new_id);
        if( icon_name.empty() )
            menu->endCascade();
        else
            menu->endCascade(const_cast<char*>(icon_name.c_str()));
    }
}


/**
   保存／復元の定型的なメニューの生成処理。
 **/
template<typename WSCMenuType>
void MakeAttributeMenuStrings(
    const std::vector<std::string>&	tag_enum,
    const std::string&              restore_menu_str,
    const std::string&              restore_ep_str,
    const std::string&              save_menu_str,
    const std::string&              save_ep_str,
    int                             new_id,
    WSCMenuType*					menu,
    std::vector<std::string>&       submenu_r_str,
    std::vector<std::string>&       submenu_s_str,
    const std::string&              restore_icon_name="",
    const std::string&              save_icon_name=""
    )
{
    MakeAttributeMenuString(tag_enum,restore_menu_str,restore_ep_str,-1,menu,submenu_r_str,
                            restore_icon_name);
    MakeAttributeMenuString(tag_enum,save_menu_str,save_ep_str,new_id,menu,submenu_s_str,
                            save_icon_name);
}

/**
   「○○の復元」の定型処理
 **/
template<typename TMenuWindow>
const std::string GetRestoringTagName(
    TMenuWindow*        			menu,
    const std::vector<std::string>&	submenu_str
    )
{
    if( menu == NULL )
        return "";
	int n = menu->getValue();
    if( n < 0 || submenu_str.size() <= static_cast<size_t>(n) )
        return "";

    std::string	item_str(submenu_str[n]);
    std::string	tag_name;
    tag_name.resize(item_str.size()+1);
    if( sscanf( item_str.c_str(), " %[^\r\n]", &tag_name[0] ) != 1 )
        return "";

    tag_name = tag_name.c_str();
    tag_name.erase(tag_name.find_last_not_of(" ")+1);
    return tag_name;
}

/**
   「○○の保存」の定型処理
 **/
template<typename TMenuWindow>
const std::string GetSavingTagName(
    TMenuWindow*        			menu,
    const std::vector<std::string>&	submenu_str,
    const std::string&		        newtag_prompt_str,
    int                         	new_item_id
    )
{
    if( menu == NULL )
        return "";
	int n = menu->getValue();

    std::string	tag_name;

    if( n == new_item_id ) {
        bool ok = true;
        do {
            std::string	input_text;
            if( InputBox( newtag_prompt_str, input_text ) != AC_OK )
                return "";

            tag_name.resize(input_text.size()+1 );
            if( sscanf( input_text.c_str()," %[^\r\n]",&tag_name[0]) != 1 )
                return "";

            std::vector<std::string>::const_iterator	it;
            for( it = submenu_str.begin(); it != submenu_str.end(); ++it ) {
                std::string	existing_name;
                existing_name = al::string::trim(*it);
                if( input_text == existing_name ) {
                    ok = false;
                    break;
                }
            }

            if( !ok )
                ok = (MsgBox( "既存の名前と重複しています。上書きしますか？", AC_YES|AC_NO) == AC_YES);

        } while( !ok );
    }
    else {
        if( n < 0 || submenu_str.size() <= static_cast<size_t>(n) )
            return "";
        std::string	item_str(submenu_str[n]);
        tag_name.resize(item_str.size()+1);
        if( sscanf( item_str.c_str(), " %[^\r\n]", &tag_name[0] ) != 1 )
            return "";
    }

    tag_name = tag_name.c_str();
    tag_name.erase(tag_name.find_last_not_of(" ")+1);
    return tag_name;
}


/** ウィンドウ配列中で最上位にあるウィンドウ（リスト中に親ウィンドウが現われないもの）のリストを返す。*/
std::vector<WSCbase*> TopWindowsInList(const std::vector<WSCbase*>& window_list);

/** 先祖兄弟従兄弟その他を含めた Window全体の家系図を作る。 */
void MakeWindowPedigree(WSCbase* window, std::set<WSCbase*>& window_list);

/** Window家系図にwindowを追加し、更にその先（親と子供）を辿る。 */
void AddWindowToPedigree(WSCbase* window, std::set<WSCbase*>& window_list);


struct ALMenuItem {
    std::string		_item_str;
    std::string		_ep_name;
    std::string     _accelerator;
    int				_item_value;
    std::string		_icon_name;
    bool            _enable;
    bool            _display;

    ALMenuItem(
        const std::string&		item_str,
        const std::string&		ep_name,
        const std::string&      accelerator,
        int						item_value,
        const std::string&		icon_name="",
        bool                    enable=true,
        bool                    display=true)
        : _item_str(item_str),
          _ep_name(ep_name),
          _accelerator(accelerator),
          _item_value(item_value),
          _icon_name(icon_name),
          _enable(enable),
          _display(display) {}
};

/** 全てのメニューの項目の文字列が一番長いものと同じ表示長になるように、
    スペースを付加する。<br>
    menu_widthが与えられている時は初期値として利用する。<br>
    all_but_spがtrueの時は、_display 対象でなくても操作する。<br>
    セパレータ（"SP"）は処理しない。<br>
    内部の文字列の操作ではアイコンの幅は必要無いが、一つでもアイコンが有効な項目
    があれば、*menu_width には反映する。
**/
template<typename TMenuWindow>
void AdjustMenuItemStr(
    TMenuWindow*        			menu,
    std::vector<ALMenuItem>&        menu_item,
    unsigned short*                 menu_width=NULL,
    bool                            all_but_sp=false
    )
{
    if( menu == NULL )
        return;

    unsigned char font_no = menu->getProperty(WSNfont);
    WSDfont* font = WSGIappFontSet()->getFont(font_no);
    const std::string	space = " ";
    size_t	space_width = GetTextWidth(space,font);
    size_t	icon_width = 0;

    size_t	max_width = (menu_width != NULL ? *menu_width : 0);
    for( size_t i = 0; i < menu_item.size(); ++i ) {
        if( menu_item[i]._item_str == "SP"
            || (!all_but_sp && !menu_item[i]._display) ) continue;
        size_t width = GetTextWidth(menu_item[i]._item_str,font);
        if( max_width < width )
            max_width = width;
        if( !menu_item[i]._icon_name.empty() )
            icon_width = 32;
    }
    max_width += space_width;	// きっちりだとメニューの右端が切れる

    if( max_width == 0 || space_width == 0 )
        return;                 /* bug! */

    for( size_t i = 0; i < menu_item.size(); ++i ) {
        if( menu_item[i]._item_str == "SP"
            || (!all_but_sp && !menu_item[i]._display) ) continue;
        
        while( GetTextWidth(menu_item[i]._item_str,font) < max_width )
            menu_item[i]._item_str += space;
    }

    if( menu_width != NULL )
        *menu_width = (max_width + icon_width);
}

/** display 対象の menu_item をmenuに addItem する。 <br>
    n_items と n_seps には追加された項目数とセパレータ数が加算される。
 **/
template<typename TMenuWindow>
void AddMenuItems(
    TMenuWindow*        			menu,
    std::vector<ALMenuItem>&        menu_item,
    int&			                n_items,
    int&            			    n_seps
    )
{
    bool	sep_allowed = false;
    int     sep_valid = false;
    for( size_t i = 0; i < menu_item.size(); ++i ) {
        if( !menu_item[i]._display ) continue;
        if( menu_item[i]._item_str == "SP" ) {
            if( sep_allowed )
                sep_valid = true;
        }
        else {
            if( sep_valid ) {
                ++n_seps;
                menu->addItem("SP","","",0);
                sep_valid = false;
            }
            ++n_items;
            menu->addItem(const_cast<char*>(menu_item[i]._item_str.c_str()),
                          const_cast<char*>(menu_item[i]._ep_name.c_str()),
                          const_cast<char*>(menu_item[i]._accelerator.c_str()),
                          menu_item[i]._item_value,
                          (menu_item[i]._icon_name.empty() ? NULL :
                           const_cast<char*>(menu_item[i]._icon_name.c_str())));
            sep_allowed = true;
        }
    }
}

/** メニュー項目に _enable 状態を反映する。 **/
template<typename TMenuWindow>
void SetMenuItemsSensitive(
    TMenuWindow*        			menu,
    const std::vector<ALMenuItem>&  menu_item
    )
{
    for( size_t i = 0; i < menu_item.size(); ++i ) {
        if( !menu_item[i]._display || menu_item[i]._item_str == "SP" ) continue;
        menu->setValueSensitive(menu_item[i]._item_value,menu_item[i]._enable);
    }
}

/** n_itemsとn_sepsからメニュー領域の高さを計算する。 **/
template<typename TMenuWindow>
unsigned short ComputeMenuHeight(
    TMenuWindow*        			menu,
    int                             n_items,
    int                             n_seps
    )
{
    if( menu == NULL )
        return 0;

    /** メニューの高さを計算する。*/
    unsigned short shadow_thickness = menu->getProperty(WSNshadowThickness);
    unsigned short separator_height = shadow_thickness*2 + 4;
    unsigned short item_height = menu->getProperty(WSNmenuItemHeight);
    unsigned short menu_height = static_cast<unsigned short>
        (shadow_thickness + n_items*item_height
         + n_seps*separator_height + shadow_thickness);
    return menu_height;
}

/** 指定されたインスタンスを始点としてフォーカスを設定出来るインスタンスを探して
    最初に見つかったものにフォーカスを設定する。
    設定出来るものが何も無ければfalseを返す。*/
bool SetFocusAround(WSCbase* object);


/** 順方向の移動（tab,downward,rightward）に関するプロパティを設定する。<br>
    現在操作が無効なものに関してはスキップする。<br>
    rotate==trueの時は先頭と末尾を繋ぐ。falseの時は繋がない。<br>
    donot_overwrite==falseの時は既にプロパティ設定されているものに関しては設定しない。
 */
void SetupForwardOrder(const std::vector<WSCbase*>& objects,
                       const std::string prop_fw,
                       bool rotate,
                       bool donot_overwrite=false);

/** 逆方向の移動（stab,upward,leftward）に関するプロパティを設定する。<br>
    現在操作が無効なものに関してはスキップする。<br>
    rotate==trueの時は先頭と末尾を繋ぐ。falseの時は繋がない。<br>
    donot_overwrite==falseの時は既にプロパティ設定されているものに関しては設定しない。
 */
void SetupBackwardOrder(const std::vector<WSCbase*>& objects,
                        const std::string prop_bk,
                        bool rotate,
                        bool donot_overwrite=false);


/** 移動に関するプロパティを設定する。<br>
    現在操作が無効なものに関してはスキップする。<br>
    rotate==trueの時は先頭と末尾を繋ぐ。falseの時は繋がない。<br>
    donot_overwrite==falseの時は既にプロパティ設定されているものに関しては設定しない。
 */
void SetupOrder(const std::vector<WSCbase*>& objects,
                const std::string prop_fw,
                const std::string prop_bw,
                bool rotate,
                bool donot_overwrite=false);


} /* namespace gui */
} /* namespace al */

#endif /*ALUTGUI_HPP_*/
